#!/usr/bin/perl -w
##############################################################################
# A perl script to work with TMT and TMTpro complementary ions in MGF files. #
# The script removes complementary ions and optionally replaces them with    #
# equivalent reporter ions.                                                  #
##############################################################################
# COPYRIGHT NOTICE                                                           #
# Copyright (C) 2022 Matrix Science Limited  All Rights Reserved.            #
#                                                                            #
# This script is free software. You can redistribute and/or                  #
# modify it under the terms of the GNU General Public License                #
# as published by the Free Software Foundation; either version 2             #
# of the License or, (at your option), any later version.                    #
#                                                                            #
# These modules are distributed in the hope that they will be useful,        #
# but WITHOUT ANY WARRANTY; without even the implied warranty of             #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the               #
# GNU General Public License for more details.                               #
#                                                                            #
# Original author: Richard Jacob                                             #
##############################################################################
use strict;
use File::Copy qw(move);
use Getopt::Std;

#Command line inputs
my (
    %opts,
    $filename,
    $upperMass,
    $replaceFlag,
    $TMTproFlag,
    $debugFlag,
    $includeDoubleChargedFlag,
);
$replaceFlag = 0;
$TMTproFlag = 0;
$TMTproFlag = 0;
$debugFlag = 0;
$includeDoubleChargedFlag = 0;

getopts("hrpdlf:", \%opts);
usage() if $opts{h};

if (defined $opts{f}){
    $filename = $opts{f};
}

if (defined $opts{r}){
    $replaceFlag = 1;
}

if (defined $opts{p}){
    $TMTproFlag = 1;
}

if (defined $opts{l}){
    $debugFlag = 1;
}

if (defined $opts{d}){
    $includeDoubleChargedFlag = 1;
}

if (not defined $filename) {
    usage();
}

my (
    $outputFilename,
    @array,
    $mz,
    @mzArray,
    $z,
    $polarity,
    $m,
    $p,
    $newQueryFlaq,
    $printFlag,
    @query,
    $line,
    $CompLine,
    @CompLineSplit,
    $originalFilename,
    $logfile,
    $MHplus,
    $TMTpro_lower,
    $TMTpro_upper,
    $TMTpro_lower2,
    $TMTpro_upper2,
    $reporterIonLower,
    $reporterIonUpper,
    $tempmass,
    $NoQueries,
    $NoCompIonsRemoved,
    $NoCompIonsRemoved2,
    $NoReporterIonsRemoved,
    $NoReporterMinusComp,
    $printCompIonsFlag,
    @compIonLines,
    @compIonLinesReverse,
    @reporterIonLines,
    @compIonLines2,
    $doubleChargedFlag,
    $CO,
    $reporter,
    $compLineReplaced,
);

#initialize variables
$p = 1.007276; #mass of proton
$TMTpro_lower = 0; 
$TMTpro_upper = 1;
$NoQueries = 0;
$NoCompIonsRemoved = 0;
$NoCompIonsRemoved2 = 0;
$NoReporterMinusComp = 0;
$NoReporterIonsRemoved = 0;
$printCompIonsFlag = 0;
$CO = 27.994915;

if ($debugFlag ==1){
    $logfile = $filename . ".TMT_complementary_ions.tsv";;
}

$outputFilename = $filename . ".out";
if ($debugFlag ==1){
    ##log for bug checking
    open (LOG, ">>$logfile") or die "could not open $logfile: $!";
    print LOG "$filename\n";
}

open (INPUT, "$filename") or die "could not open $filename: $!";
open (OUTPUT, ">$outputFilename") or die "could not open $outputFilename: $!";
$newQueryFlaq = 0;

if ($TMTproFlag==1) {
    $reporterIonLower = 126;
    $reporterIonUpper = 135.5;
}
else{
    $reporterIonLower = 126;
    $reporterIonUpper = 131.5;
}
if ($debugFlag ==1){
    print LOG "M\\Z\tMH+\tZ\tLower\tUpper\tLower2\tUpper2\tNo reporter ions\tNo Comp ions\tNo Comp ions doubly charged\tdoubleChargedFlag\tReporter minus Comp ions\n";
}
while (<INPUT>){
    #if (/\#/) {
    #    #header line
    #    print OUTPUT $_;
    #    next;
    #}
    #if (/_/) {
    #    #header line
    #    print OUTPUT $_;
    #    next;
    #}
    
    if (/BEGIN IONS/){
        $newQueryFlaq = 1;
        $printCompIonsFlag = 1;
        $doubleChargedFlag = 0,
        $TMTpro_lower2 = 0;
        $TMTpro_upper2 = 0;
        $NoQueries++;
        push (@query, $_);
        next;
    }
    if (/PEPMASS/){
        #PEPMASS=638.86504 6207476
        @mzArray = split(/[\s\=]/, $_);
        $mz = $mzArray[1];
        #($mz) =/(\d*\.\d*)/;
        push (@query, $_);
        next;
    }
    if (/CHARGE/){
        #CHARGE=1+
        ($z) =/(\d)/;
        #if chage state greater than 2 then the complementary ions will probably be doubly charged
        if ($includeDoubleChargedFlag ==1){
            if ($z>2) {
                $doubleChargedFlag = 1,
            }
        }
        ($polarity) = /([+-])/;
        if ($polarity eq "+"){
           $m =  ($mz * $z)-(($z-1) * $p);
        }
        else{
            $m =  ($mz * $z)+(($z-1) * $p);
        }
        push (@query, $_);
        #precusor mass - CO (28Da) - reporter ion. Use largest and smallest reporter ions for range. Nominal mass is fine.
        if ($TMTproFlag==1) {
            $TMTpro_lower = $m-136-28;
            $TMTpro_upper = $m-125-28;
            if ($doubleChargedFlag == 1){
                $TMTpro_lower2 = (($m-136-28+$p)/2);
                $TMTpro_upper2 = (($m-125-28+$p)/2);
            }
        }
        else {
            $TMTpro_lower = $m-131.5-28;
            $TMTpro_upper = $m-125-28;
            if ($doubleChargedFlag == 1){
                $TMTpro_lower2 = (($m-131.5-28+$p)/2);
                $TMTpro_upper2 = (($m-125-28+$p)/2);
            }
        }
        if ($debugFlag ==1){
            print LOG "$mz\t$m\t$z\t$TMTpro_lower\t$TMTpro_upper\t$TMTpro_lower2\t$TMTpro_upper2\t";
        }
        next;
    }
    if ($newQueryFlaq == 1){
        if (/END IONS/){
            if ($doubleChargedFlag==1) {
                $NoReporterMinusComp = $NoReporterIonsRemoved - $NoCompIonsRemoved2;
            }
            else{
                $NoReporterMinusComp = $NoReporterIonsRemoved - $NoCompIonsRemoved;
            }

            if ($debugFlag ==1){
                print LOG "$NoReporterIonsRemoved\t$NoCompIonsRemoved\t$NoCompIonsRemoved2\t$doubleChargedFlag\t$NoReporterMinusComp\n";
            }
            $NoReporterIonsRemoved= 0;
            $NoCompIonsRemoved = 0;
            $NoCompIonsRemoved2 = 0;
            $newQueryFlaq = 0;
            $tempmass = '';
            push @query, $_;
            foreach $line (@query){
                ($tempmass) = ($line=~/(^\d*\.\d*)/);
                if (!defined $tempmass || $tempmass eq ''){
                    print OUTPUT $line;
                    next;
                }
                if ($tempmass >= $reporterIonUpper and $printCompIonsFlag==1 and $replaceFlag==1) {
                    if ($doubleChargedFlag == 1){
                        #Check to see how many ions are doubly charged compared to singly charged. If there are more than than singly charged then output them instead
                        if (scalar @compIonLines2 >= scalar @compIonLines) {
                            @compIonLines = @compIonLines2;
                        }                        
                    }
                    #reverse the complementary ions to print them in the correct order in the peak list
                    @compIonLinesReverse = reverse @compIonLines;
                    foreach $CompLine (@compIonLinesReverse){
                        print OUTPUT "$CompLine\n";
                    }
                    $printCompIonsFlag=0;
                }
                print OUTPUT $line;
            }
            @query = ();
            @compIonLines = ();
            @compIonLines2 = ();
            @reporterIonLines = ();
            next;
        }
        ($tempmass) =/(^\d*\.\d*)/;
        if (!defined $tempmass || $tempmass eq ''){
            push @query, $_;
            next;
        }
        if ($tempmass >= $TMTpro_lower  and  $tempmass <= $TMTpro_upper) {
            $NoCompIonsRemoved ++;
            (@CompLineSplit) = split (/\s/, $_);
            # reporterIon = MH+-( CompIon + CO + H)
            $reporter = $m -($CompLineSplit[0]+$CO - $p);
            $CompLineSplit[0] = $reporter;
            $compLineReplaced = join( ' ', @CompLineSplit );
            push @compIonLines, $compLineReplaced;
            next;
        }
        if ($tempmass >= $TMTpro_lower2  and  $tempmass <= $TMTpro_upper2) {
            $NoCompIonsRemoved2 ++;
            (@CompLineSplit) = split (/\s/, $_);
            # reporterIon = MH+-( CompIon + CO + H)
            #account for double charge
            $reporter = $m -((($CompLineSplit[0] * 2)-$p) +$CO - $p);
            $CompLineSplit[0] = $reporter;
            $compLineReplaced = join( ' ', @CompLineSplit );
            push @compIonLines2, $compLineReplaced;
            next;
        }
                
        if ($tempmass >= $reporterIonLower  and  $tempmass <= $reporterIonUpper) {
            $NoReporterIonsRemoved ++;
            push @reporterIonLines, $_;
            next;
        }
        else{
            push @query, $_;
        }
    }
}
close INPUT;
close OUTPUT;

if ($debugFlag ==1){
    print LOG "No Queries processed\t$NoQueries\tNo Comp ions removed\t$NoCompIonsRemoved\n";
    close LOG;
}

if ($NoQueries == 0) {
    warn "Input file had zero 'BEGIN IONS' lines. Is it really an MGF file? Aborting";
    exit;
}

$originalFilename = $filename . ".orig";
if ($debugFlag ==1){
    rename ($filename, $originalFilename)|| die ( "Error in renaming original file: $!" );
}
unlink $filename;
rename ($outputFilename, $filename)|| die ( "Error in renaming out file: $!" );;
unlink $outputFilename;


sub usage {
print STDERR << "eof";

TMT_complementary_ions.pl -- Removes complementary ions from peak list and
optionally replaces them with reporter ions.

Usage:
$0 [-r] [-p] [-d] [-l] -f filename

    -h              : this (help) message
    -f filename     : Input data filename
    -r              : Output reporter ions equivalent to removed complementary ions
    -p              : TMT by default, or TMTpro if -p specified
    -d              : Include doubly charged complementary ions
    -l              : Logging mode records a log and saves the original peaklist

The script overwrites the input file. When the logging mode is active, the
original file is saved in the same location as the input file with the file
extension ".orig".

Command line example:

perl TMT_complementary_ions.pl -f "F1.mgf" -r -p -d -l

Mascot Daemon example:

perl "\\path\\to\\TMT_complementary_ions.pl" -f "<cachedpeaklist>" -r -p -d -l
eof
    exit;
}

